@@ -1,5 +1,5 @@
 /* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
-/* $Id: memcached.h,v 1.21 2004/02/24 23:42:02 bradfitz Exp $ */
+/* $Id: memcached.h 393 2006-09-09 22:12:07Z bradfitz $ */
 
 #define DATA_BUFFER_SIZE 2048
 
@@ -24,11 +24,12 @@ struct stats {
 };
 
 struct settings {
-    unsigned int maxbytes;
+    size_t maxbytes;
     int maxconns;
     int port;
     struct in_addr interface;
     int verbose;
+    int managed;          /* if 1, a tracker manages virtual buckets */
     time_t oldest_live;   /* ignore existing items older than this */
     int evict_to_free;
 };
@@ -83,11 +84,12 @@ typedef struct {
     int    state;
     struct event event;
     short  ev_flags;
-    short  which;  /* which events were just triggered */
+    short  which;   /* which events were just triggered */
 
-    char   *rbuf;  
-    int    rsize;  
-    int    rbytes;
+    char   *rbuf;   /* buffer to read commands into */
+    char   *rcurr;  /* but if we parsed some already, this is where we stopped */
+    int    rsize;   /* total allocated size of rbuf */
+    int    rbytes;  /* how much data, starting from rcur, do we have unparsed */
 
     char   *wbuf;
     char   *wcurr;
@@ -97,7 +99,7 @@ typedef struct {
     void   *write_and_free; /* free this memory after finishing writing */
     char    is_corked;         /* boolean, connection is corked */
 
-    char   *rcurr;
+    char   *ritem;  /* when we read in an item's value, it goes here */
     int    rlbytes;
     
     /* data for the nread state */
@@ -123,9 +125,16 @@ typedef struct {
     char   ibuf[300]; /* for VALUE lines */
     char   *iptr;
     int    ibytes;
+    int    binary;    /* are we in binary mode */
+    int    bucket;    /* bucket number for the next command, if running as
+                         a managed instance. -1 (_not_ 0) means invalid. */
+    int    gen;       /* generation requested for the bucket */
                          
 } conn;
 
+/* number of virtual buckets for a managed instance */
+#define MAX_BUCKETS 32768
+
 /* listening socket */
 extern int l_socket;
 
@@ -148,17 +157,25 @@ time_t realtime(time_t exptime);
 /* slabs memory allocation */
 
 /* Init the subsystem. The argument is the limit on no. of bytes to allocate, 0 if no limit */
-void slabs_init(unsigned int limit);
+void slabs_init(size_t limit);
+
+/* Preallocate as many slab pages as possible (called from slabs_init)
+   on start-up, so users don't get confused out-of-memory errors when
+   they do have free (in-slab) space, but no space to make new slabs.
+   if maxslabs is 18 (POWER_LARGEST - POWER_SMALLEST + 1), then all
+   slab types can be made.  if max memory is less than 18 MB, only the
+   smaller ones will be made.  */
+void slabs_preallocate (unsigned int maxslabs);
 
 /* Given object size, return id to use when allocating/freeing memory for object */
 /* 0 means error: can't store such a large object */
-unsigned int slabs_clsid(unsigned int size);
+unsigned int slabs_clsid(size_t size);
 
 /* Allocate object of given length. 0 on error */
-void *slabs_alloc(unsigned int size);
+void *slabs_alloc(size_t size);
 
 /* Free previously allocated object */
-void slabs_free(void *ptr, unsigned int size);
+void slabs_free(void *ptr, size_t size);
     
 /* Fill buffer with stats */
 char* slabs_stats(int *buflen);
