@@ -2,7 +2,7 @@
 /*
  * Slabs memory allocation, based on powers-of-2
  *
- * $Id: slabs.c,v 1.15 2003/09/05 22:37:36 bradfitz Exp $
+ * $Id: slabs.c 393 2006-09-09 22:12:07Z bradfitz $
  */
 
 #include <sys/types.h>
@@ -49,10 +49,10 @@ typedef struct {
 } slabclass_t;
 
 static slabclass_t slabclass[POWER_LARGEST+1];
-static unsigned int mem_limit = 0;
-static unsigned int mem_malloced = 0;
+static size_t mem_limit = 0;
+static size_t mem_malloced = 0;
 
-unsigned int slabs_clsid(unsigned int size) {
+unsigned int slabs_clsid(size_t size) {
     int res = 1;
 
     if(size==0)
@@ -67,7 +67,7 @@ unsigned int slabs_clsid(unsigned int size) {
     return res;
 }
 
-void slabs_init(unsigned int limit) {
+void slabs_init(size_t limit) {
     int i;
     int size=1;
 
@@ -83,6 +83,41 @@ void slabs_init(unsigned int limit) {
         slabclass[i].list_size = 0;
         slabclass[i].killing = 0;
     }
+
+    /* for the test suite:  faking of how much we've already malloc'd */
+    {
+        char *t_initial_malloc = getenv("T_MEMD_INITIAL_MALLOC");
+        if (t_initial_malloc) {
+            mem_malloced = atol(getenv("T_MEMD_INITIAL_MALLOC"));
+        }
+    }
+
+    /* pre-allocate slabs by default, unless the environment variable
+       for testing is set to something non-zero */
+    {
+        char *pre_alloc = getenv("T_MEMD_SLABS_ALLOC");
+        if (!pre_alloc || atoi(pre_alloc)) {
+            slabs_preallocate(limit / POWER_BLOCK);
+        }
+    }
+}
+
+void slabs_preallocate (unsigned int maxslabs) {
+    int i;
+    unsigned int prealloc = 0;
+
+    /* pre-allocate a 1MB slab in every size class so people don't get
+       confused by non-intuitive "SERVER_ERROR out of memory"
+       messages.  this is the most common question on the mailing
+       list.  if you really don't want this, you can rebuild without
+       these three lines.  */
+
+    for(i=POWER_SMALLEST; i<=POWER_LARGEST; i++) {
+        if (++prealloc > maxslabs)
+            return;
+        slabs_newslab(i);
+    } 
+ 
 }
 
 static int grow_slab_list (unsigned int id) { 
@@ -120,7 +155,7 @@ int slabs_newslab(unsigned int id) {
     return 1;
 }
 
-void *slabs_alloc(unsigned int size) {
+void *slabs_alloc(size_t size) {
     slabclass_t *p;
 
     unsigned char id = slabs_clsid(size);
@@ -160,7 +195,7 @@ void *slabs_alloc(unsigned int size) {
     return 0;  /* shouldn't ever get here */
 }
 
-void slabs_free(void *ptr, unsigned int size) {
+void slabs_free(void *ptr, size_t size) {
     unsigned char id = slabs_clsid(size);
     slabclass_t *p;
 
@@ -216,7 +251,7 @@ char* slabs_stats(int *buflen) {
             total++;
         }
     }
-    bufcurr += sprintf(bufcurr, "STAT active_slabs %d\r\nSTAT total_malloced %u\r\n", total, mem_malloced);
+    bufcurr += sprintf(bufcurr, "STAT active_slabs %d\r\nSTAT total_malloced %llu\r\n", total, (unsigned long long) mem_malloced);
     bufcurr += sprintf(bufcurr, "END\r\n");
     *buflen = bufcurr - buf;
     return buf;
